"""Module that groups built-in connectors for mail2beyond to use."""
from . import void
from . import smtp
from . import slack
from . import discord
from . import google_chat
from . import microsoft_teams
